# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name, import-error, no-name-in-module
"""
This file contains the majority of the migration tests that are too short to
go to a separate file.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from six.moves import range

import numpy

from aiida.backends.testbase import AiidaTestCase
from aiida.backends.djsite.db.migrations import ModelModifierV0025
from aiida.backends.general.migrations import utils
from aiida.common.exceptions import IntegrityError
from aiida.manage.database.integrity.duplicate_uuid import deduplicate_uuids, verify_uuid_uniqueness
from .test_migrations_common import TestMigrations


class TestMigrationsModelModifierV0025(TestMigrations):
    """Sub class of `TestMigrations` that need to work on node attributes using the `ModelModifierV0025`."""

    def set_attribute(self, node, key, value):
        DbAttribute = self.apps.get_model('db', 'DbAttribute')
        modifier = ModelModifierV0025(self.apps, DbAttribute)
        modifier.set_value_for_node(node.pk, key, value)

    def get_attribute(self, node, key, default=None):  # pylint: disable=missing-docstring
        DbAttribute = self.apps.get_model('db', 'DbAttribute')
        modifier = ModelModifierV0025(self.apps, DbAttribute)
        try:
            return modifier.get_value_for_node(node.pk, key)
        except AttributeError:
            return default

    @staticmethod
    def get_node_array(node, name):
        return utils.load_numpy_array_from_repository(node.uuid, name)

    def set_node_array(self, node, name, array):
        """Store a new numpy array inside a node. Possibly overwrite the array if it already existed.

        Internally, it stores a name.npy file in numpy format.

        :param name: The name of the array.
        :param array: The numpy array to store.
        """
        utils.store_numpy_array_in_repository(node.uuid, name, array)
        self.set_attribute(node, 'array|{}'.format(name), list(array.shape))


class TestNoMigrations(AiidaTestCase):
    """Verify that no django migrations remain."""

    def test_no_remaining_migrations(self):  # pylint: disable=no-self-use
        """
        Verify that no django migrations remain.
        Equivalent to python manage.py makemigrations --check
        """

        from django.core.management import call_command

        # Raises SystemExit, if migrations remain
        call_command('makemigrations', '--check', verbosity=0)


class TestDuplicateNodeUuidMigration(TestMigrations):
    """Test the migration that verifies that there are no duplicate UUIDs"""

    migrate_from = '0013_django_1_8'
    migrate_to = '0014_add_node_uuid_unique_constraint'

    def setUpBeforeMigration(self):
        from aiida.common.utils import get_new_uuid
        self.file_name = 'test.temp'
        self.file_content = u'#!/bin/bash\n\necho test run\n'

        self.nodes_boolean = []
        self.nodes_integer = []
        self.n_bool_duplicates = 2
        self.n_int_duplicates = 4

        node_bool = self.DbNode(type='data.bool.Bool.', user_id=self.default_user.id, uuid=get_new_uuid())
        node_bool.save()

        node_int = self.DbNode(type='data.int.Int.', user_id=self.default_user.id, uuid=get_new_uuid())
        node_int.save()

        self.nodes_boolean.append(node_bool)
        self.nodes_integer.append(node_int)

        for _ in range(self.n_bool_duplicates):
            node = self.DbNode(type='data.bool.Bool.', user_id=self.default_user.id, uuid=node_bool.uuid)
            node.save()
            utils.put_object_from_string(node.uuid, self.file_name, self.file_content)
            self.nodes_boolean.append(node)

        for _ in range(self.n_int_duplicates):
            node = self.DbNode(type='data.int.Int.', user_id=self.default_user.id, uuid=node_int.uuid)
            node.save()
            utils.put_object_from_string(node.uuid, self.file_name, self.file_content)
            self.nodes_integer.append(node)

        # Verify that there are duplicate UUIDs by checking that the following function raises
        with self.assertRaises(IntegrityError):
            verify_uuid_uniqueness(table='db_dbnode')

        # Now run the function responsible for solving duplicate UUIDs which would also be called by the user
        # through the `verdi database integrity detect-duplicate-uuid` command
        deduplicate_uuids(table='db_dbnode', dry_run=False)

    def test_deduplicated_uuids(self):
        """Verify that after the migration, all expected nodes are still there with unique UUIDs."""
        # If the duplicate UUIDs were successfully fixed, the following should not raise.
        verify_uuid_uniqueness(table='db_dbnode')

        # Reload the nodes by PK and check that all UUIDs are now unique
        nodes_boolean = [self.load_node(node.pk) for node in self.nodes_boolean]
        uuids_boolean = [node.uuid for node in nodes_boolean]
        self.assertEqual(len(set(uuids_boolean)), len(nodes_boolean))

        nodes_integer = [self.load_node(node.pk) for node in self.nodes_integer]
        uuids_integer = [node.uuid for node in nodes_integer]
        self.assertEqual(len(set(uuids_integer)), len(nodes_integer))

        for node in nodes_boolean:
            self.assertEqual(utils.get_object_from_repository(node.uuid, self.file_name), self.file_content)


class TestUuidMigration(TestMigrations):
    """
    This test class checks the migration 0018_django_1_11 which switches from the django_extensions
    UUID field to the native UUIDField of django 1.11. It also introduces unique constraints
    on all uuid columns (previously existed only on dbnode).
    """

    migrate_from = '0017_drop_dbcalcstate'
    migrate_to = '0018_django_1_11'

    def setUpBeforeMigration(self):
        node = self.DbNode(type='node.process.calculation.calcjob.CalcJobNode.', user_id=self.default_user.id)
        node.save()

        self.node_uuid = str(node.uuid)
        self.node_id = node.id

    def test_uuid_untouched(self):
        """Verify that Node uuids remain unchanged."""
        node = self.load_node(self.node_id)
        self.assertEqual(self.node_uuid, str(node.uuid))


class TestGroupRenamingMigration(TestMigrations):
    """
    This test class checks the migration 0022_dbgroup_type_string_change_content which updates the
    type_string column of the groups.
    """

    migrate_from = '0021_dbgroup_name_to_label_type_to_type_string'
    migrate_to = '0022_dbgroup_type_string_change_content'

    def setUpBeforeMigration(self):
        # Create group
        DbGroup = self.apps.get_model('db', 'DbGroup')

        # test user group type_string: '' -> 'user'
        group_user = DbGroup(label='test_user_group', user_id=self.default_user.id, type_string='')
        group_user.save()
        self.group_user_pk = group_user.pk

        # test data.upf group type_string: 'data.upf.family' -> 'data.upf'
        group_data_upf = DbGroup(
            label='test_data_upf_group', user_id=self.default_user.id, type_string='data.upf.family'
        )
        group_data_upf.save()
        self.group_data_upf_pk = group_data_upf.pk

        # test auto.import group type_string: 'aiida.import' -> 'auto.import'
        group_autoimport = DbGroup(label='test_import_group', user_id=self.default_user.id, type_string='aiida.import')
        group_autoimport.save()
        self.group_autoimport_pk = group_autoimport.pk

        # test auto.run group type_string: 'autogroup.run' -> 'auto.run'
        group_autorun = DbGroup(label='test_autorun_group', user_id=self.default_user.id, type_string='autogroup.run')
        group_autorun.save()
        self.group_autorun_pk = group_autorun.pk

    def test_group_string_update(self):
        """ Test that the type_string were updated correctly """
        DbGroup = self.apps.get_model('db', 'DbGroup')

        # test user group type_string: '' -> 'user'
        group_user = DbGroup.objects.get(pk=self.group_user_pk)
        self.assertEqual(group_user.type_string, 'user')

        # test data.upf group type_string: 'data.upf.family' -> 'data.upf'
        group_data_upf = DbGroup.objects.get(pk=self.group_data_upf_pk)
        self.assertEqual(group_data_upf.type_string, 'data.upf')

        # test auto.import group type_string: 'aiida.import' -> 'auto.import'
        group_autoimport = DbGroup.objects.get(pk=self.group_autoimport_pk)
        self.assertEqual(group_autoimport.type_string, 'auto.import')

        # test auto.run group type_string: 'autogroup.run' -> 'auto.run'
        group_autorun = DbGroup.objects.get(pk=self.group_autorun_pk)
        self.assertEqual(group_autorun.type_string, 'auto.run')


class TestCalcAttributeKeysMigration(TestMigrationsModelModifierV0025):
    """
    This test class checks that the migration 0023_calc_job_option_attribute_keys works as expected
    which migrates CalcJobNode attributes for metadata options whose key changed.
    """

    migrate_from = '0022_dbgroup_type_string_change_content'
    migrate_to = '0023_calc_job_option_attribute_keys'

    KEY_RESOURCES_OLD = 'jobresource_params'
    KEY_RESOURCES_NEW = 'resources'
    KEY_PARSER_NAME_OLD = 'parser'
    KEY_PARSER_NAME_NEW = 'parser_name'
    KEY_PROCESS_LABEL_OLD = '_process_label'
    KEY_PROCESS_LABEL_NEW = 'process_label'
    KEY_ENVIRONMENT_VARIABLES_OLD = 'custom_environment_variables'
    KEY_ENVIRONMENT_VARIABLES_NEW = 'environment_variables'

    def setUpBeforeMigration(self):
        self.process_label = 'TestLabel'
        self.resources = {'number_machines': 1}
        self.environment_variables = {}
        self.parser_name = 'aiida.parsers:parser'

        self.node_work = self.DbNode(type='node.process.workflow.WorkflowNode.', user_id=self.default_user.id)
        self.node_work.save()
        self.set_attribute(self.node_work, self.KEY_PROCESS_LABEL_OLD, self.process_label)

        self.node_calc = self.DbNode(type='node.process.calculation.calcjob.CalcJobNode.', user_id=self.default_user.id)
        self.node_calc.save()
        self.set_attribute(self.node_calc, self.KEY_PROCESS_LABEL_OLD, self.process_label)
        self.set_attribute(self.node_calc, self.KEY_RESOURCES_OLD, self.resources)
        self.set_attribute(self.node_calc, self.KEY_ENVIRONMENT_VARIABLES_OLD, self.environment_variables)
        self.set_attribute(self.node_calc, self.KEY_PARSER_NAME_OLD, self.parser_name)

        # Create a node of a different type to ensure that its attributes are not updated
        self.node_other = self.DbNode(type='node.othernode.', user_id=self.default_user.id)
        self.node_other.save()
        self.set_attribute(self.node_other, self.KEY_PROCESS_LABEL_OLD, self.process_label)
        self.set_attribute(self.node_other, self.KEY_RESOURCES_OLD, self.resources)
        self.set_attribute(self.node_other, self.KEY_ENVIRONMENT_VARIABLES_OLD, self.environment_variables)
        self.set_attribute(self.node_other, self.KEY_PARSER_NAME_OLD, self.parser_name)

    def test_attribute_key_changes(self):
        """Verify that the keys are successfully changed of the affected attributes."""
        NOT_FOUND = tuple([0])

        self.assertEqual(self.get_attribute(self.node_work, self.KEY_PROCESS_LABEL_NEW), self.process_label)
        self.assertEqual(self.get_attribute(self.node_work, self.KEY_PROCESS_LABEL_OLD, default=NOT_FOUND), NOT_FOUND)

        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_PROCESS_LABEL_NEW), self.process_label)
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_RESOURCES_NEW), self.resources)
        self.assertEqual(
            self.get_attribute(self.node_calc, self.KEY_ENVIRONMENT_VARIABLES_NEW), self.environment_variables
        )
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_PARSER_NAME_NEW), self.parser_name)
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_PROCESS_LABEL_OLD, default=NOT_FOUND), NOT_FOUND)
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_RESOURCES_OLD, default=NOT_FOUND), NOT_FOUND)
        self.assertEqual(
            self.get_attribute(self.node_calc, self.KEY_ENVIRONMENT_VARIABLES_OLD, default=NOT_FOUND), NOT_FOUND
        )
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_PARSER_NAME_OLD, default=NOT_FOUND), NOT_FOUND)

        # The following node should not be migrated even if its attributes have the matching keys because
        # the node is not a ProcessNode
        self.assertEqual(self.get_attribute(self.node_other, self.KEY_PROCESS_LABEL_OLD), self.process_label)
        self.assertEqual(self.get_attribute(self.node_other, self.KEY_RESOURCES_OLD), self.resources)
        self.assertEqual(
            self.get_attribute(self.node_other, self.KEY_ENVIRONMENT_VARIABLES_OLD), self.environment_variables
        )
        self.assertEqual(self.get_attribute(self.node_other, self.KEY_PARSER_NAME_OLD), self.parser_name)
        self.assertEqual(self.get_attribute(self.node_other, self.KEY_PROCESS_LABEL_NEW, default=NOT_FOUND), NOT_FOUND)
        self.assertEqual(self.get_attribute(self.node_other, self.KEY_RESOURCES_NEW, default=NOT_FOUND), NOT_FOUND)
        self.assertEqual(
            self.get_attribute(self.node_other, self.KEY_ENVIRONMENT_VARIABLES_NEW, default=NOT_FOUND), NOT_FOUND
        )
        self.assertEqual(self.get_attribute(self.node_other, self.KEY_PARSER_NAME_NEW, default=NOT_FOUND), NOT_FOUND)


class TestDbLogMigrationRecordCleaning(TestMigrations):
    """
    This test class checks that the migration 0024_dblog_update works as expected.
    That migration updates of the DbLog table and adds uuids
    """

    migrate_from = '0023_calc_job_option_attribute_keys'
    migrate_to = '0024_dblog_update'

    def setUpBeforeMigration(self):  # pylint: disable=too-many-locals
        import json
        import importlib
        from aiida.backends.general.migrations.utils import dumps_json

        update_024 = importlib.import_module('aiida.backends.djsite.db.migrations.0024_dblog_update')

        DbNode = self.apps.get_model('db', 'DbNode')
        DbWorkflow = self.apps.get_model('db', 'DbWorkflow')
        DbLog = self.apps.get_model('db', 'DbLog')

        # Creating the needed nodes & workflows
        calc_1 = DbNode(type='node.process.calculation.CalculationNode.', user_id=self.default_user.id)
        param = DbNode(type='data.dict.Dict.', user_id=self.default_user.id)
        leg_workf = DbWorkflow(label='Legacy WorkflowNode', user_id=self.default_user.id)
        calc_2 = DbNode(type='node.process.calculation.CalculationNode.', user_id=self.default_user.id)

        # Storing them
        calc_1.save()
        param.save()
        leg_workf.save()
        calc_2.save()

        # Creating the corresponding log records and storing them
        log_1 = DbLog(
            loggername='CalculationNode logger',
            objpk=calc_1.pk,
            objname='node.calculation.job.quantumespresso.pw.',
            message='calculation node 1',
            metadata=json.dumps({
                'msecs': 719.0849781036377,
                'objpk': calc_1.pk,
                'lineno': 350,
                'thread': 140011612940032,
                'asctime': '10/21/2018 12:39:51 PM',
                'created': 1540118391.719085,
                'levelno': 23,
                'message': 'calculation node 1',
                'objname': 'node.calculation.job.quantumespresso.pw.',
            })
        )
        log_2 = DbLog(
            loggername='something.else logger',
            objpk=param.pk,
            objname='something.else.',
            message='parameter data with log message'
        )
        log_3 = DbLog(
            loggername='TopologicalWorkflow logger',
            objpk=leg_workf.pk,
            objname='aiida.workflows.user.topologicalworkflows.topo.TopologicalWorkflow',
            message='parameter data with log message'
        )
        log_4 = DbLog(
            loggername='CalculationNode logger',
            objpk=calc_2.pk,
            objname='node.calculation.job.quantumespresso.pw.',
            message='calculation node 2',
            metadata=json.dumps({
                'msecs': 719.0849781036377,
                'objpk': calc_2.pk,
                'lineno': 360,
                'levelno': 23,
                'message': 'calculation node 1',
                'objname': 'node.calculation.job.quantumespresso.pw.',
            })
        )
        # Creating two more log records that don't correspond to a node
        log_5 = DbLog(
            loggername='CalculationNode logger',
            objpk=(calc_2.pk + 1000),
            objname='node.calculation.job.quantumespresso.pw.',
            message='calculation node 1000',
            metadata=json.dumps({
                'msecs': 718,
                'objpk': (calc_2.pk + 1000),
                'lineno': 361,
                'levelno': 25,
                'message': 'calculation node 1000',
                'objname': 'node.calculation.job.quantumespresso.pw.',
            })
        )
        log_6 = DbLog(
            loggername='CalculationNode logger',
            objpk=(calc_2.pk + 1001),
            objname='node.calculation.job.quantumespresso.pw.',
            message='calculation node 10001',
            metadata=json.dumps({
                'msecs': 722,
                'objpk': (calc_2.pk + 1001),
                'lineno': 362,
                'levelno': 24,
                'message': 'calculation node 1001',
                'objname': 'node.calculation.job.quantumespresso.pw.',
            })
        )

        # Storing the log records
        log_1.save()
        log_2.save()
        log_3.save()
        log_4.save()
        log_5.save()
        log_6.save()

        # Storing temporarily information needed for the check at the test
        self.to_check = dict()

        # Keeping calculation & calculation log ids
        self.to_check['CalculationNode'] = (
            calc_1.pk,
            log_1.pk,
            calc_2.pk,
            log_4.pk,
        )

        # Getting the serialized Dict logs
        param_data = DbLog.objects.filter(objpk=param.pk).filter(objname='something.else.'
                                                                ).values(*update_024.values_to_export)[:1]
        serialized_param_data = dumps_json(list(param_data))
        # Getting the serialized logs for the unknown entity logs (as the export migration fuction
        # provides them) - this should coincide to the above
        serialized_unknown_exp_logs = update_024.get_serialized_unknown_entity_logs(self.schema_editor)
        # Getting their number
        unknown_exp_logs_number = update_024.get_unknown_entity_log_number(self.schema_editor)
        self.to_check['Dict'] = (serialized_param_data, serialized_unknown_exp_logs, unknown_exp_logs_number)

        # Getting the serialized legacy workflow logs
        leg_wf = DbLog.objects.filter(objpk=leg_workf.pk).filter(
            objname='aiida.workflows.user.topologicalworkflows.topo.TopologicalWorkflow'
        ).values(*update_024.values_to_export)[:1]
        serialized_leg_wf_logs = dumps_json(list(leg_wf))
        # Getting the serialized logs for the legacy workflow logs (as the export migration function
        # provides them) - this should coincide to the above
        serialized_leg_wf_exp_logs = update_024.get_serialized_legacy_workflow_logs(self.schema_editor)
        eg_wf_exp_logs_number = update_024.get_legacy_workflow_log_number(self.schema_editor)
        self.to_check['WorkflowNode'] = (serialized_leg_wf_logs, serialized_leg_wf_exp_logs, eg_wf_exp_logs_number)

        # Getting the serialized logs that don't correspond to a DbNode record
        logs_no_node = DbLog.objects.filter(id__in=[log_5.id, log_6.id]).values(*update_024.values_to_export)
        serialized_logs_no_node = dumps_json(list(logs_no_node))
        # Getting the serialized logs that don't correspond to a node (as the export migration function
        # provides them) - this should coincide to the above
        serialized_logs_exp_no_node = update_024.get_serialized_logs_with_no_nodes(self.schema_editor)
        logs_no_node_number = update_024.get_logs_with_no_nodes_number(self.schema_editor)
        self.to_check['NoNode'] = (serialized_logs_no_node, serialized_logs_exp_no_node, logs_no_node_number)

    def tearDown(self):
        """Cleaning the DbLog, DbUser, DbWorkflow and DbNode records"""
        DbUser = self.apps.get_model('db', 'DbUser')
        DbNode = self.apps.get_model('db', 'DbNode')
        DbWorkflow = self.apps.get_model('db', 'DbWorkflow')
        DbLog = self.apps.get_model('db', 'DbLog')

        DbLog.objects.all().delete()
        DbNode.objects.all().delete()  # pylint: disable=no-member
        DbWorkflow.objects.all().delete()  # pylint: disable=no-member
        DbUser.objects.all().delete()  # pylint: disable=no-member
        super(TestDbLogMigrationRecordCleaning, self).tearDown()

    def test_dblog_calculation_node(self):
        """
        Verify that after the migration there is only two log records left and verify that they corresponds to
        the CalculationNodes.
        """
        DbLog = self.apps.get_model('db', 'DbLog')

        # Check that only two log records exist
        self.assertEqual(DbLog.objects.count(), 2, 'There should be two log records left')

        # Get the node id of the log record referencing the node and verify that it is the correct one
        dbnode_id_1 = DbLog.objects.filter(pk=self.to_check['CalculationNode'][1]
                                          ).values('dbnode_id')[:1].get()['dbnode_id']
        self.assertEqual(dbnode_id_1, self.to_check['CalculationNode'][0], 'referenced node is not the expected one')
        dbnode_id_2 = DbLog.objects.filter(pk=self.to_check['CalculationNode'][3]
                                          ).values('dbnode_id')[:1].get()['dbnode_id']
        self.assertEqual(dbnode_id_2, self.to_check['CalculationNode'][2], 'referenced node is not the expected one')

    def test_dblog_correct_export_of_logs(self):
        """
        Verify that export log methods for legacy workflows, unknown entities and log records that
        don't correspond to nodes, work as expected
        """
        import json

        self.assertEqual(self.to_check['Dict'][0], self.to_check['Dict'][1])
        self.assertEqual(self.to_check['Dict'][2], 1)

        self.assertEqual(self.to_check['WorkflowNode'][0], self.to_check['WorkflowNode'][1])
        self.assertEqual(self.to_check['WorkflowNode'][2], 1)

        self.assertEqual(
            sorted(list(json.loads(self.to_check['NoNode'][0])), key=lambda k: k['id']),
            sorted(list(json.loads(self.to_check['NoNode'][1])), key=lambda k: k['id'])
        )
        self.assertEqual(self.to_check['NoNode'][2], 2)

    def test_dblog_unique_uuids(self):
        """
        Verify that the UUIDs of the log records are unique
        """
        DbLog = self.apps.get_model('db', 'DbLog')

        l_uuids = list(_['uuid'] for _ in DbLog.objects.values('uuid'))
        s_uuids = set(l_uuids)
        self.assertEqual(len(l_uuids), len(s_uuids), 'The UUIDs are not all unique.')

    def test_metadata_correctness(self):
        """
        Verify that the metadata of the remaining records don't have an objpk and objmetadata values.
        """
        import json

        DbLog = self.apps.get_model('db', 'DbLog')

        metadata = list(json.loads(_['metadata']) for _ in DbLog.objects.values('metadata'))
        # Verify that the objpk and objname are no longer part of the metadata
        for m_res in metadata:
            self.assertNotIn('objpk', m_res.keys(), 'objpk should not exist any more in metadata')
            self.assertNotIn('objname', m_res.keys(), 'objname should not exist any more in metadata')


class TestDbLogMigrationBackward(TestMigrations):
    """
    Check that backward migrations work also for the DbLog migration(s).
    """

    migrate_from = '0024_dblog_update'
    migrate_to = '0023_calc_job_option_attribute_keys'

    def setUpBeforeMigration(self):
        import json

        DbNode = self.apps.get_model('db', 'DbNode')
        DbLog = self.apps.get_model('db', 'DbLog')

        # Creating the needed nodes & workflows
        calc_1 = DbNode(type='node.process.calculation.CalculationNode.1', user_id=self.default_user.id)
        calc_2 = DbNode(type='node.process.calculation.CalculationNode.2', user_id=self.default_user.id)

        # Storing them
        calc_1.save()
        calc_2.save()

        # Creating the corresponding log records and storing them
        log_1 = DbLog(
            loggername='CalculationNode logger',
            dbnode_id=calc_1.pk,
            message='calculation node 1',
            metadata=json.dumps({
                'msecs': 719.0849781036377,
                'lineno': 350,
                'thread': 140011612940032,
                'asctime': '10/21/2018 12:39:51 PM',
                'created': 1540118391.719085,
                'levelno': 23,
                'message': 'calculation node 1',
            })
        )
        log_2 = DbLog(
            loggername='CalculationNode logger',
            dbnode_id=calc_2.pk,
            message='calculation node 2',
            metadata=json.dumps({
                'msecs': 719.0849781036377,
                'lineno': 360,
                'levelno': 23,
                'message': 'calculation node 1',
            })
        )

        # Storing the log records
        log_1.save()
        log_2.save()

        # Keeping what is needed to be verified at the test
        self.to_check = dict()
        self.to_check[log_1.pk] = (log_1.dbnode_id, calc_1.type)
        self.to_check[log_2.pk] = (log_2.dbnode_id, calc_2.type)

    def test_objpk_objname(self):
        """
        This test verifies that the objpk and objname have the right values
        after a forward and a backward migration.
        """
        import json
        DbLog = self.apps.get_model('db', 'DbLog')

        # Check that only two log records exist with the correct objpk objname
        for log_pk, to_check_value in self.to_check.items():
            log_entry = DbLog.objects.filter(pk=log_pk)[:1].get()
            log_dbnode_id, node_type = to_check_value
            self.assertEqual(
                log_dbnode_id, log_entry.objpk,
                'The dbnode_id ({}) of the 0024 schema version should be identical to the objpk ({}) of '
                'the 0023 schema version.'.format(log_dbnode_id, log_entry.objpk)
            )
            self.assertEqual(
                node_type, log_entry.objname,
                'The type ({}) of the linked node of the 0024 schema version should be identical to the '
                'objname ({}) of the 0023 schema version.'.format(node_type, log_entry.objname)
            )
            self.assertEqual(
                log_dbnode_id,
                json.loads(log_entry.metadata)['objpk'],
                'The dbnode_id ({}) of the 0024 schema version should be identical to the objpk ({}) of '
                'the 0023 schema version stored in the metadata.'.format(
                    log_dbnode_id,
                    json.loads(log_entry.metadata)['objpk']
                )
            )
            self.assertEqual(
                node_type,
                json.loads(log_entry.metadata)['objname'],
                'The type ({}) of the linked node of the 0024 schema version should be identical to the '
                'objname ({}) of the 0023 schema version stored in the metadata.'.format(
                    node_type,
                    json.loads(log_entry.metadata)['objname']
                )
            )


class TestDataMoveWithinNodeMigration(TestMigrations):
    """
    Check that backward migrations work also for the DbLog migration(s).
    """

    migrate_from = '0024_dblog_update'
    migrate_to = '0025_move_data_within_node_module'

    def setUpBeforeMigration(self):
        self.node_calc = self.DbNode(type='node.process.calculation.calcjob.CalcJobNode.', user_id=self.default_user.id)
        self.node_data = self.DbNode(type='data.int.Int.', user_id=self.default_user.id)
        self.node_calc.save()
        self.node_data.save()

    def test_data_type_string(self):
        """Verify that type string of the Data node was successfully adapted."""
        node_calc = self.load_node(self.node_calc.id)
        node_data = self.load_node(self.node_data.id)
        self.assertEqual(node_data.type, 'node.data.int.Int.')
        self.assertEqual(node_calc.type, 'node.process.calculation.calcjob.CalcJobNode.')


class TestTrajectoryDataMigration(TestMigrationsModelModifierV0025):
    """
    This test class checks that the migrations 0026_trajectory_symbols_to_attribute and
    0027_delete_trajectory_symbols_array work as expected.
    These are data migrations for `TrajectoryData` nodes where symbol lists are moved
    from repository array to attributes.
    """

    migrate_from = '0025_move_data_within_node_module'
    migrate_to = '0027_delete_trajectory_symbols_array'

    stepids = numpy.array([60, 70])
    times = stepids * 0.01
    positions = numpy.array([[[0., 0., 0.], [0.5, 0.5, 0.5], [1.5, 1.5, 1.5]],
                             [[0., 0., 0.], [0.5, 0.5, 0.5], [1.5, 1.5, 1.5]]])
    velocities = numpy.array([[[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]],
                              [[0.5, 0.5, 0.5], [0.5, 0.5, 0.5], [-0.5, -0.5, -0.5]]])
    cells = numpy.array([[[2., 0., 0.], [0., 2., 0.], [0., 0., 2.]], [[3., 0., 0.], [0., 3., 0.], [0., 0., 3.]]])

    def setUpBeforeMigration(self):
        symbols = numpy.array(['H', 'O', 'C'])
        self.node = self.DbNode(type='node.data.array.trajectory.TrajectoryData.', user_id=self.default_user.id)
        self.node.save()

        self.set_node_array(self.node, 'steps', self.stepids)
        self.set_node_array(self.node, 'cells', self.cells)
        self.set_node_array(self.node, 'symbols', symbols)
        self.set_node_array(self.node, 'positions', self.positions)
        self.set_node_array(self.node, 'times', self.times)
        self.set_node_array(self.node, 'velocities', self.velocities)

    def test_trajectory_symbols(self):
        """ Check that the trajectories are migrated correctly """
        node = self.load_node(self.node.id)
        self.assertSequenceEqual(self.get_attribute(node, 'symbols'), ['H', 'O', 'C'])
        self.assertSequenceEqual(self.get_node_array(node, 'velocities').tolist(), self.velocities.tolist())
        self.assertSequenceEqual(self.get_node_array(node, 'positions').tolist(), self.positions.tolist())
        with self.assertRaises(IOError):
            self.get_node_array(node, 'symbols')


class TestNodePrefixRemovalMigration(TestMigrations):
    """
    This test class checks that the migration 0028_remove_node_prefix works as expected.

    That is the final data migration for `Nodes` after `aiida.orm.nodes` reorganization
    was finalized to remove the `node.` prefix
    """

    migrate_from = '0027_delete_trajectory_symbols_array'
    migrate_to = '0028_remove_node_prefix'

    def setUpBeforeMigration(self):
        self.node_calc = self.DbNode(type='node.process.calculation.calcjob.CalcJobNode.', user_id=self.default_user.id)
        self.node_data = self.DbNode(type='node.data.int.Int.', user_id=self.default_user.id)
        self.node_calc.save()
        self.node_data.save()

    def test_data_node_type_string(self):
        """Verify that type string of the nodes was successfully adapted."""
        node_calc = self.load_node(self.node_calc.id)
        node_data = self.load_node(self.node_data.id)

        self.assertEqual(node_data.type, 'data.int.Int.')
        self.assertEqual(node_calc.type, 'process.calculation.calcjob.CalcJobNode.')


class TestParameterDataToDictMigration(TestMigrations):
    """
    This test class checks that the migration 0029_rename_parameter_data_to_dict works as expected.

    This is a data migration for the renaming of `ParameterData` to `Dict`.
    """

    migrate_from = '0028_remove_node_prefix'
    migrate_to = '0029_rename_parameter_data_to_dict'

    def setUpBeforeMigration(self):
        self.node = self.DbNode(type='data.parameter.ParameterData.', user_id=self.default_user.id)
        self.node.save()

    def test_data_node_type_string(self):
        """Verify that type string of the nodes was successfully adapted."""
        node = self.load_node(self.node.id)
        self.assertEqual(node.type, 'data.dict.Dict.')


class TestTextFieldToJSONFieldMigration(TestMigrations):  # pylint: disable=too-many-instance-attributes
    """
    This test class checks that the migration 0033_replace_text_field_with_json_field works as expected.

    That migration replaces the use of text fields to store JSON data with builtin JSONFields.
    """

    migrate_from = '0032_remove_legacy_workflows'
    migrate_to = '0033_replace_text_field_with_json_field'

    def setUpBeforeMigration(self):
        from aiida.common import json

        self.DbNode = self.apps.get_model('db', 'DbNode')
        self.DbComputer = self.apps.get_model('db', 'DbComputer')
        self.DbAuthInfo = self.apps.get_model('db', 'DbAuthInfo')
        self.DbLog = self.apps.get_model('db', 'DbLog')

        self.node = self.DbNode(node_type='node.process.calculation.CalculationNode.', user_id=self.default_user.id)
        self.node.save()

        self.computer_metadata = {
            'shebang': '#!/bin/bash',
            'workdir': '/scratch/',
            'append_text': '',
            'prepend_text': '',
            'mpirun_command': ['mpirun', '-np', '{tot_num_mpiprocs}'],
            'default_mpiprocs_per_machine': 1
        }
        self.computer_kwargs = {
            'name': 'localhost_testing',
            'hostname': 'localhost',
            'transport_type': 'local',
            'scheduler_type': 'direct',
            'metadata': json.dumps(self.computer_metadata),
        }
        self.computer = self.DbComputer(**self.computer_kwargs)
        self.computer.save()

        self.auth_info_auth_params = {'safe_interval': 2}
        self.auth_info_metadata = {'safe_interval': 2}
        self.auth_info_kwargs = {
            'aiidauser_id': self.default_user.pk,
            'dbcomputer': self.computer,
            'auth_params': json.dumps(self.auth_info_auth_params),
            'metadata': json.dumps(self.auth_info_metadata),
        }
        self.auth_info = self.DbAuthInfo(**self.auth_info_kwargs)
        self.auth_info.save()

        self.log_metadata = {
            'msecs': 719.0849781036377,
            'lineno': 350,
            'thread': 140011612940032,
            'asctime': '10/21/2018 12:39:51 PM',
            'created': 1540118391.719085,
            'levelno': 23,
            'message': 'calculation node 1',
        }
        self.log_kwargs = {
            'loggername': 'localhost',
            'levelname': 'localhost',
            'dbnode_id': self.node.id,
            'metadata': json.dumps(self.log_metadata)
        }
        self.log = self.DbLog(**self.log_kwargs)
        self.log.save()

    def test_text_field_to_json_field_migration(self):
        """Verify that the values in the text fields were maintained after migrating the field to JSONField."""
        # Reload the objects to make sure the new data is loaded
        computer = self.DbComputer.objects.get(pk=self.computer.id)
        auth_info = self.DbAuthInfo.objects.get(pk=self.auth_info.id)
        log = self.DbLog.objects.get(pk=self.log.id)

        # Make sure that the migrated data matches the original
        self.assertDictEqual(computer.metadata, self.computer_metadata)
        self.assertDictEqual(auth_info.metadata, self.auth_info_metadata)
        self.assertDictEqual(auth_info.auth_params, self.auth_info_auth_params)
        self.assertDictEqual(log.metadata, self.log_metadata)


class TestResetHash(TestMigrations):
    """
    This test class checks that only the hash extra is removed.
    """

    migrate_from = '0038_data_migration_legacy_job_calculations'
    migrate_to = '0039_reset_hash'

    def setUpBeforeMigration(self):
        self.node = self.DbNode(
            node_type='process.calculation.calcjob.CalcJobNode.',
            user_id=self.default_user.id,
            extras={
                'something': 123,
                '_aiida_hash': 'abcd'
            }
        )
        self.node.save()

    def test_data_migrated(self):
        """Verify that type string of the nodes was successfully adapted."""
        node = self.load_node(self.node.id)
        extras = node.extras
        self.assertEqual(extras.get('something'), 123)  # Other extras should be untouched
        self.assertNotIn('_aiida_hash', extras)  # The hash extra should have been removed



#!/usr/bin/env python
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=inconsistent-return-statements
"""
It defines the method with all required parameters to run restapi locally.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import imp
import os

from flask_cors import CORS


def run_api(flask_app, flask_api, **kwargs):
    """
    Takes a flask.Flask instance and runs it.

    flask_app: Class inheriting from Flask app class
    flask_api = flask_restful API class to be used to wrap the app

    kwargs:
    List of valid parameters:
    prog_name: name of the command before arguments are parsed. Useful when
    api is embedded in a command, such as verdi restapi
    hostname: self-explainatory
    port: self-explainatory
    config: directory containing the config.py file used to
    configure the RESTapi
    catch_internal_server: If true, catch and print all inter server errors
    debug: self-explainatory
    wsgi_profile:to use WSGI profiler middleware for finding bottlenecks in web application
    hookup: to hookup app
    All other passed parameters are ignored.
    """
    # pylint: disable=too-many-locals

    # Unpack parameters
    hostname = kwargs['hostname']
    port = kwargs['port']
    config = kwargs['config']

    catch_internal_server = kwargs.pop('catch_internal_server', False)
    debug = kwargs['debug']
    wsgi_profile = kwargs['wsgi_profile']
    hookup = kwargs['hookup']

    # Import the right configuration file
    confs = imp.load_source(os.path.join(config, 'config'), os.path.join(config, 'config.py'))

    # Instantiate an app
    app_kwargs = dict(catch_internal_server=catch_internal_server)
    app = flask_app(__name__, **app_kwargs)

    # Config the app
    app.config.update(**confs.APP_CONFIG)

    # cors
    cors_prefix = os.path.join(confs.PREFIX, '*')
    CORS(app, resources={r'' + cors_prefix: {'origins': '*'}})

    # Config the serializer used by the app
    if confs.SERIALIZER_CONFIG:
        from aiida.restapi.common.utils import CustomJSONEncoder
        app.json_encoder = CustomJSONEncoder

    # If the user selects the profiling option, then we need
    # to do a little extra setup
    if wsgi_profile:
        from werkzeug.contrib.profiler import ProfilerMiddleware

        app.config['PROFILE'] = True
        app.wsgi_app = ProfilerMiddleware(app.wsgi_app, restrictions=[30])

    # Instantiate an Api by associating its app
    api_kwargs = dict(PREFIX=confs.PREFIX, PERPAGE_DEFAULT=confs.PERPAGE_DEFAULT, LIMIT_DEFAULT=confs.LIMIT_DEFAULT)
    api = flask_api(app, **api_kwargs)

    # Check if the app has to be hooked-up or just returned
    if hookup:
        print(' * REST API running on http://{}:{}{}'.format(hostname, port, confs.PREFIX))
        api.app.run(debug=debug, host=hostname, port=int(port), threaded=True)

    else:
        # here we return the app, and the api with no specifications on debug
        #  mode, port and host. This can be handled by an external server,
        # e.g. apache2, which will set the host and port. This implies that
        # the user-defined configuration of the app is ineffective (it only
        # affects the internal werkzeug server used by Flask).
        return (app, api)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""The main `verdi` click group."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from aiida.cmdline.params import options, types


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@options.PROFILE(type=types.ProfileParamType(load_profile=True))
@click.version_option(None, '-v', '--version', message='AiiDA version %(version)s')
@click.pass_context
def verdi(ctx, profile):
    """The command line interface of AiiDA."""
    from aiida.common import extendeddicts
    from aiida.manage.configuration import get_config

    if ctx.obj is None:
        ctx.obj = extendeddicts.AttributeDict()

    ctx.obj.config = get_config()
    ctx.obj.profile = profile



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for comment backend classes."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import abc
import six

from . import backends

__all__ = ('BackendComment', 'BackendCommentCollection')


@six.add_metaclass(abc.ABCMeta)
class BackendComment(backends.BackendEntity):
    """Base class for a node comment."""

    @property
    def uuid(self):
        return str(self._dbmodel.uuid)

    @abc.abstractproperty
    def ctime(self):
        pass

    @abc.abstractproperty
    def mtime(self):
        pass

    @abc.abstractmethod
    def set_mtime(self, value):
        pass

    @abc.abstractproperty
    def node(self):
        pass

    @abc.abstractproperty
    def user(self):
        pass

    @abc.abstractmethod
    def set_user(self, value):
        pass

    @abc.abstractproperty
    def content(self):
        pass

    @abc.abstractmethod
    def set_content(self, value):
        pass


@six.add_metaclass(abc.ABCMeta)
class BackendCommentCollection(backends.BackendCollection[BackendComment]):
    """The collection of Comment entries."""

    ENTITY_CLASS = BackendComment

    @abc.abstractmethod
    def create(self, node, user, content=None, **kwargs):
        """
        Create a Comment for a given node and user

        :param node: a Node instance
        :param user: a User instance
        :param content: the comment content
        :return: a Comment object associated to the given node and user
        """

    @abc.abstractmethod
    def delete(self, comment_id):
        """
        Remove a Comment from the collection with the given id

        :param comment_id: the id of the comment to delete
        :type comment_id: int

        :raises TypeError: if ``comment_id`` is not an `int`
        :raises `~aiida.common.exceptions.NotExistent`: if Comment with ID ``comment_id`` is not found
        """

    @abc.abstractmethod
    def delete_all(self):
        """
        Delete all Comment entries.

        :raises `~aiida.common.exceptions.IntegrityError`: if all Comments could not be deleted
        """

    @abc.abstractmethod
    def delete_many(self, filters):
        """
        Delete Comments based on ``filters``

        :param filters: similar to QueryBuilder filter
        :type filters: dict

        :return: (former) ``PK`` s of deleted Comments
        :rtype: list

        :raises TypeError: if ``filters`` is not a `dict`
        :raises `~aiida.common.exceptions.ValidationError`: if ``filters`` is empty
        """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=import-error,no-name-in-module,invalid-name
""" The basic functionality for the migration tests"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from django.apps import apps
from django.db.migrations.executor import MigrationExecutor
from django.db import connection

from aiida.backends.testbase import AiidaTestCase
from aiida.common.utils import Capturing


class TestMigrations(AiidaTestCase):
    """
    This is the common test class that is used by all migration tests. It migrates to a given
    migration point, allows you to set up the database & AiiDA at that point with the necessary
    data and migrates then to the final migration point.
    In the end it forwards the database at the final migration (as it should be and found before
    the migration tests).
    """

    @property
    def app(self):
        return apps.get_containing_app_config(type(self).__module__).name.split('.')[-1]

    migrate_from = None
    migrate_to = None

    def setUp(self):
        """Go to a specific schema version before running tests."""
        from aiida.backends import sqlalchemy as sa
        from aiida.orm import autogroup

        self.current_autogroup = autogroup.current_autogroup
        autogroup.current_autogroup = None
        assert self.migrate_from and self.migrate_to, \
            "TestCase '{}' must define migrate_from and migrate_to properties".format(type(self).__name__)
        self.migrate_from = [(self.app, self.migrate_from)]
        self.migrate_to = [(self.app, self.migrate_to)]
        executor = MigrationExecutor(connection)
        self.apps = executor.loader.project_state(self.migrate_from).apps
        self.schema_editor = connection.schema_editor()

        # Reset session for the migration
        sa.get_scoped_session().close()
        # Reverse to the original migration
        with Capturing():
            executor.migrate(self.migrate_from)
        # Reset session after the migration
        sa.get_scoped_session().close()

        self.DbNode = self.apps.get_model('db', 'DbNode')
        self.DbUser = self.apps.get_model('db', 'DbUser')
        self.DbUser.objects.all().delete()
        self.default_user = self.DbUser(1, 'aiida@localhost')
        self.default_user.save()

        try:
            self.setUpBeforeMigration()
            # Run the migration to test
            executor = MigrationExecutor(connection)
            executor.loader.build_graph()

            # Reset session for the migration
            sa.get_scoped_session().close()
            with Capturing():
                executor.migrate(self.migrate_to)
            # Reset session after the migration
            sa.get_scoped_session().close()

            self.apps = executor.loader.project_state(self.migrate_to).apps
        except Exception:
            # Bring back the DB to the correct state if this setup part fails
            import traceback
            traceback.print_stack()
            self._revert_database_schema()
            raise

    def tearDown(self):
        """At the end make sure we go back to the latest schema version."""
        from aiida.orm import autogroup
        self._revert_database_schema()
        autogroup.current_autogroup = self.current_autogroup

    def setUpBeforeMigration(self):
        """Anything to do before running the migrations, which should be implemented in test subclasses."""

    def _revert_database_schema(self):
        """Bring back the DB to the correct state."""
        from ...migrations import LATEST_MIGRATION
        from aiida.backends import sqlalchemy as sa

        self.migrate_to = [(self.app, LATEST_MIGRATION)]

        # Reset session for the migration
        sa.get_scoped_session().close()
        executor = MigrationExecutor(connection)
        with Capturing():
            executor.migrate(self.migrate_to)
        # Reset session after the migration
        sa.get_scoped_session().close()

    def load_node(self, pk):
        return self.DbNode.objects.get(pk=pk)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Migration from v0.3 to v0.4, used by `verdi export migrate` command.

The migration steps are named similarly to the database migrations for Django and SQLAlchemy.
In the description of each migration, a revision number is given, which refers to the Django migrations.
The individual Django database migrations may be found at:

    `aiida.backends.djsite.db.migrations.00XX_<migration-name>.py`

Where XX are the numbers in the migrations' documentation: REV. 1.0.XX
And migration-name is the name of the particular migration.
The individual SQLAlchemy database migrations may be found at:

    `aiida.backends.sqlalchemy.migrations.versions.<id>_<migration-name>.py`

Where id is a SQLA id and migration-name is the name of the particular migration.
"""
# pylint: disable=invalid-name
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os

import numpy as np

from aiida.cmdline.utils import echo
from aiida.tools.importexport.migration.utils import verify_metadata_version, update_metadata, remove_fields


def migration_base_data_plugin_type_string(data):
    """Apply migration: 0009 - REV. 1.0.9
    `DbNode.type` content changes:
    'data.base.Bool.'  -> 'data.bool.Bool.'
    'data.base.Float.' -> 'data.float.Float.'
    'data.base.Int.'   -> 'data.int.Int.'
    'data.base.Str.'   -> 'data.str.Str.'
    'data.base.List.'  -> 'data.list.List.'
    """
    for content in data['export_data'].get('Node', {}).values():
        if content.get('type', '').startswith('data.base.'):
            type_str = content['type'].replace('data.base.', '')
            type_str = 'data.' + type_str.lower() + type_str
            content['type'] = type_str


def migration_process_type(metadata, data):
    """Apply migrations: 0010 - REV. 1.0.10
    Add `DbNode.process_type` column
    """
    # For data.json
    for content in data['export_data'].get('Node', {}).values():
        if 'process_type' not in content:
            content['process_type'] = ''
    # For metadata.json
    metadata['all_fields_info']['Node']['process_type'] = {}


def migration_code_sub_class_of_data(data):
    """Apply migrations: 0016 - REV. 1.0.16
    The Code class used to be just a sub class of Node, but was changed to act like a Data node.
    code.Code. -> data.code.Code.
    """
    for content in data['export_data'].get('Node', {}).values():
        if content.get('type', '') == 'code.Code.':
            content['type'] = 'data.code.Code.'


def migration_add_node_uuid_unique_constraint(data):
    """Apply migration: 0014 - REV. 1.0.14, 0018 - REV. 1.0.18
    Check that no entries with the same uuid are present in the export file
    if yes - stop the import process
    """
    for entry_type in ['Group', 'Computer', 'Node']:
        if entry_type not in data['export_data']:  # if a particular entry type is not present - skip
            continue
        all_uuids = [content['uuid'] for content in data['export_data'][entry_type].values()]
        unique_uuids = set(all_uuids)
        if len(all_uuids) != len(unique_uuids):
            echo.echo_critical(
                """{}s with exactly the same UUID found, cannot proceed further. Please contact AiiDA
            developers: http://www.aiida.net/mailing-list/ to help you resolve this issue.""".format(entry_type)
            )


def migration_migrate_builtin_calculations(data):
    """Apply migrations: 0019 - REV. 1.0.19
    Remove 'simpleplugin' from ArithmeticAddCalculation and TemplatereplacerCalculation type

    ATTENTION:

    The 'process_type' column did not exist before migration 0010, consequently, it could not be present in any
    export archive of the currently existing stable releases (0.12.*). Here, however, the migration acts
    on the content of the 'process_type' column, which could only be introduced in alpha releases of AiiDA 1.0.
    Assuming that 'add' and 'templateplacer' calculations are expected to have both 'type' and 'process_type' columns,
    they will be added based solely on the 'type' column content (unlike the way it is done in the DB migration,
    where the 'type_string' content was also checked).
    """
    for key, content in data['export_data'].get('Node', {}).items():
        if content.get('type', '') == 'calculation.job.simpleplugins.arithmetic.add.ArithmeticAddCalculation.':
            content['type'] = 'calculation.job.arithmetic.add.ArithmeticAddCalculation.'
            content['process_type'] = 'aiida.calculations:arithmetic.add'
        elif content.get('type', '') == 'calculation.job.simpleplugins.templatereplacer.TemplatereplacerCalculation.':
            content['type'] = 'calculation.job.templatereplacer.TemplatereplacerCalculation.'
            content['process_type'] = 'aiida.calculations:templatereplacer'
        elif content.get('type', '') == 'data.code.Code.':
            if data['node_attributes'][key]['input_plugin'] == 'simpleplugins.arithmetic.add':
                data['node_attributes'][key]['input_plugin'] = 'arithmetic.add'

            elif data['node_attributes'][key]['input_plugin'] == 'simpleplugins.templatereplacer':
                data['node_attributes'][key]['input_plugin'] = 'templatereplacer'


def migration_provenance_redesign(data):  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    """Apply migration: 0020 - REV. 1.0.20
    Provenance redesign
    """
    from aiida.manage.database.integrity.plugins import infer_calculation_entry_point
    from aiida.manage.database.integrity import write_database_integrity_violation
    from aiida.plugins.entry_point import ENTRY_POINT_STRING_SEPARATOR

    fallback_cases = []
    calcjobs_to_migrate = {}

    for key, value in data['export_data'].get('Node', {}).items():
        if value.get('type', '').startswith('calculation.job.'):
            calcjobs_to_migrate[key] = value

    if calcjobs_to_migrate:
        # step1: rename the type column of process nodes
        mapping_node_entry = infer_calculation_entry_point(
            type_strings=[e['type'] for e in calcjobs_to_migrate.values()]
        )
        for uuid, content in calcjobs_to_migrate.items():
            type_string = content['type']
            entry_point_string = mapping_node_entry[type_string]

            # If the entry point string does not contain the entry point string separator,
            # the mapping function was not able to map the type string onto a known entry point string.
            # As a fallback it uses the modified type string itself.
            # All affected entries should be logged to file that the user can consult.
            if ENTRY_POINT_STRING_SEPARATOR not in entry_point_string:
                fallback_cases.append([uuid, type_string, entry_point_string])

            content['process_type'] = entry_point_string

        if fallback_cases:
            headers = ['UUID', 'type (old)', 'process_type (fallback)']
            warning_message = 'found calculation nodes with a type string ' \
                              'that could not be mapped onto a known entry point'
            action_message = 'inferred `process_type` for all calculation nodes, ' \
                             'using fallback for unknown entry points'
            write_database_integrity_violation(fallback_cases, headers, warning_message, action_message)

    # step2: detect and delete unexpected links
    action_message = 'the link was deleted'
    headers = ['UUID source', 'UUID target', 'link type', 'link label']

    def delete_wrong_links(node_uuids, link_type, headers, warning_message, action_message):
        """delete links that are matching link_type and are going from nodes listed in node_uuids"""
        violations = []
        new_links_list = []
        for link in data['links_uuid']:
            if link['input'] in node_uuids and link['type'] == link_type:
                violations.append([link['input'], link['output'], link['type'], link['label']])
            else:
                new_links_list.append(link)
        data['links_uuid'] = new_links_list
        if violations:
            write_database_integrity_violation(violations, headers, warning_message, action_message)

    # calculations with outgoing CALL links
    calculation_uuids = {
        value['uuid'] for value in data['export_data'].get('Node', {}).values() if (
            value.get('type', '').startswith('calculation.job.') or
            value.get('type', '').startswith('calculation.inline.')
        )
    }
    warning_message = 'detected calculation nodes with outgoing `call` links.'
    delete_wrong_links(calculation_uuids, 'calllink', headers, warning_message, action_message)

    # calculations with outgoing RETURN links
    warning_message = 'detected calculation nodes with outgoing `return` links.'
    delete_wrong_links(calculation_uuids, 'returnlink', headers, warning_message, action_message)

    # outgoing CREATE links from FunctionCalculation and WorkCalculation nodes
    warning_message = 'detected outgoing `create` links from FunctionCalculation and/or WorkCalculation nodes.'
    work_uuids = {
        value['uuid'] for value in data['export_data'].get('Node', {}).values() if (
            value.get('type', '').startswith('calculation.function') or
            value.get('type', '').startswith('calculation.work')
        )
    }
    delete_wrong_links(work_uuids, 'createlink', headers, warning_message, action_message)

    for node_id, node in data['export_data'].get('Node', {}).items():
        # migrate very old `ProcessCalculation` to `WorkCalculation`
        if node.get('type', '') == 'calculation.process.ProcessCalculation.':
            node['type'] = 'calculation.work.WorkCalculation.'

        #  WorkCalculations that have a `function_name` attribute are FunctionCalculations
        if node.get('type', '') == 'calculation.work.WorkCalculation.':
            if (
                'function_name' in data['node_attributes'][node_id] and
                data['node_attributes'][node_id]['function_name'] is not None
            ):
                # for some reason for the workchains the 'function_name' attribute is present but has None value
                node['type'] = 'node.process.workflow.workfunction.WorkFunctionNode.'
            else:
                node['type'] = 'node.process.workflow.workchain.WorkChainNode.'

        # update type for JobCalculation nodes
        if node.get('type', '').startswith('calculation.job.'):
            node['type'] = 'node.process.calculation.calcjob.CalcJobNode.'

        # update type for InlineCalculation nodes
        if node.get('type', '') == 'calculation.inline.InlineCalculation.':
            node['type'] = 'node.process.calculation.calcfunction.CalcFunctionNode.'

        # update type for FunctionCalculation nodes
        if node.get('type', '') == 'calculation.function.FunctionCalculation.':
            node['type'] = 'node.process.workflow.workfunction.WorkFunctionNode.'

    uuid_node_type_mapping = {
        node['uuid']: node['type'] for node in data['export_data'].get('Node', {}).values() if 'type' in node
    }
    for link in data['links_uuid']:
        inp_uuid = link['output']
        # rename `createlink` to `create`
        if link['type'] == 'createlink':
            link['type'] = 'create'
        # rename `returnlink` to `return`
        elif link['type'] == 'returnlink':
            link['type'] = 'return'

        elif link['type'] == 'inputlink':
            # rename `inputlink` to `input_calc` if the target node is a calculation type node
            if uuid_node_type_mapping[inp_uuid].startswith('node.process.calculation'):
                link['type'] = 'input_calc'
            # rename `inputlink` to `input_work` if the target node is a workflow type node
            elif uuid_node_type_mapping[inp_uuid].startswith('node.process.workflow'):
                link['type'] = 'input_work'

        elif link['type'] == 'calllink':
            # rename `calllink` to `call_calc` if the target node is a calculation type node
            if uuid_node_type_mapping[inp_uuid].startswith('node.process.calculation'):
                link['type'] = 'call_calc'
            # rename `calllink` to `call_work` if the target node is a workflow type node
            elif uuid_node_type_mapping[inp_uuid].startswith('node.process.workflow'):
                link['type'] = 'call_work'


def migration_dbgroup_name_to_label_type_to_type_string(metadata, data):
    """Apply migrations: 0021 - REV. 1.0.21
    Rename dbgroup fields:
    name -> label
    type -> type_string
    """
    # For data.json
    for content in data['export_data'].get('Group', {}).values():
        if 'name' in content:
            content['label'] = content.pop('name')
        if 'type' in content:
            content['type_string'] = content.pop('type')
    # For metadata.json
    metadata_group = metadata['all_fields_info']['Group']
    if 'name' in metadata_group:
        metadata_group['label'] = metadata_group.pop('name')
    if 'type' in metadata_group:
        metadata_group['type_string'] = metadata_group.pop('type')


def migration_dbgroup_type_string_change_content(data):
    """Apply migrations: 0022 - REV. 1.0.22
    Change type_string according to the following rule:
    '' -> 'user'
    'data.upf.family' -> 'data.upf'
    'aiida.import' -> 'auto.import'
    'autogroup.run' -> 'auto.run'
    """
    for content in data['export_data'].get('Group', {}).values():
        key_mapper = {
            '': 'user',
            'data.upf.family': 'data.upf',
            'aiida.import': 'auto.import',
            'autogroup.run': 'auto.run'
        }
        if content['type_string'] in key_mapper:
            content['type_string'] = key_mapper[content['type_string']]


def migration_calc_job_option_attribute_keys(data):
    """Apply migrations: 0023 - REV. 1.0.23
    `custom_environment_variables` -> `environment_variables`
    `jobresource_params` -> `resources`
    `_process_label` -> `process_label`
    `parser` -> `parser_name`
    """

    # Helper function
    def _migration_calc_job_option_attribute_keys(attr_id, content):
        """Apply migration 0023 - REV. 1.0.23 for both `node_attributes*` dicts in `data.json`"""
        # For CalcJobNodes only
        if data['export_data']['Node'][attr_id]['type'] == 'node.process.calculation.calcjob.CalcJobNode.':
            key_mapper = {
                'custom_environment_variables': 'environment_variables',
                'jobresource_params': 'resources',
                'parser': 'parser_name'
            }
            for key in content:
                if key in key_mapper:
                    content[key_mapper[key]] = content.pop(key)

        # For all processes
        if data['export_data']['Node'][attr_id]['type'].startswith('node.process.'):
            if '_process_label' in content:
                content['process_label'] = content.pop('_process_label')

    # Update node_attributes and node_attributes_conversion
    attribute_dicts = ['node_attributes', 'node_attributes_conversion']
    for attribute_dict in attribute_dicts:
        for attr_id, content in data[attribute_dict].items():
            if 'type' in data['export_data'].get('Node', {}).get(attr_id, {}):
                _migration_calc_job_option_attribute_keys(attr_id, content)


def migration_move_data_within_node_module(data):
    """Apply migrations: 0025 - REV. 1.0.25
    The type string for `Data` nodes changed from `data.*` to `node.data.*`.
    """
    for value in data['export_data'].get('Node', {}).values():
        if value.get('type', '').startswith('data.'):
            value['type'] = value['type'].replace('data.', 'node.data.', 1)


def migration_trajectory_symbols_to_attribute(data, folder):
    """Apply migrations: 0026 - REV. 1.0.26 and 0027 - REV. 1.0.27
    Create the symbols attribute from the repository array for all `TrajectoryData` nodes.
    """
    from aiida.tools.importexport.common.config import NODES_EXPORT_SUBFOLDER

    for node_id, content in data['export_data'].get('Node', {}).items():
        if content.get('type', '') == 'node.data.array.trajectory.TrajectoryData.':
            uuid = content['uuid']
            symbols_path = os.path.join(
                folder.get_abs_path(NODES_EXPORT_SUBFOLDER), uuid[0:2], uuid[2:4], uuid[4:], 'path', 'symbols.npy'
            )
            symbols = np.load(symbols_path).tolist()
            os.remove(symbols_path)
            # Update 'node_attributes'
            data['node_attributes'][node_id].pop('array|symbols', None)
            data['node_attributes'][node_id]['symbols'] = symbols
            # Update 'node_attributes_conversion'
            data['node_attributes_conversion'][node_id].pop('array|symbols', None)
            data['node_attributes_conversion'][node_id]['symbols'] = [None] * len(symbols)


def migration_remove_node_prefix(data):
    """Apply migrations: 0028 - REV. 1.0.28
    Change node type strings:
    'node.data.' -> 'data.'
    'node.process.' -> 'process.'
    """
    for value in data['export_data'].get('Node', {}).values():
        if value.get('type', '').startswith('node.data.'):
            value['type'] = value['type'].replace('node.data.', 'data.', 1)
        elif value.get('type', '').startswith('node.process.'):
            value['type'] = value['type'].replace('node.process.', 'process.', 1)


def migration_rename_parameter_data_to_dict(data):
    """Apply migration: 0029 - REV. 1.0.29
    Update ParameterData to Dict
    """
    for value in data['export_data'].get('Node', {}).values():
        if value.get('type', '') == 'data.parameter.ParameterData.':
            value['type'] = 'data.dict.Dict.'


def migration_dbnode_type_to_dbnode_node_type(metadata, data):
    """Apply migration: 0030 - REV. 1.0.30
    Renaming DbNode.type to DbNode.node_type
    """
    # For data.json
    for content in data['export_data'].get('Node', {}).values():
        if 'type' in content:
            content['node_type'] = content.pop('type')
    # For metadata.json
    if 'type' in metadata['all_fields_info']['Node']:
        metadata['all_fields_info']['Node']['node_type'] = metadata['all_fields_info']['Node'].pop('type')


def migration_remove_dbcomputer_enabled(metadata, data):
    """Apply migration: 0031 - REV. 1.0.31
    Remove DbComputer.enabled
    """
    remove_fields(metadata, data, ['Computer'], ['enabled'])


def migration_replace_text_field_with_json_field(data):
    """Apply migration 0033 - REV. 1.0.33
    Store dict-values as JSON serializable dicts instead of strings
    NB! Specific for Django backend
    """
    import six
    from aiida.common import json
    for content in data['export_data'].get('Computer', {}).values():
        for value in ['metadata', 'transport_params']:
            if isinstance(content[value], six.text_type):
                content[value] = json.loads(content[value])
    for content in data['export_data'].get('Log', {}).values():
        if isinstance(content['metadata'], six.text_type):
            content['metadata'] = json.loads(content['metadata'])


def add_extras(data):
    """Update data.json with the new Extras
    Since Extras were not available previously and usually only include hashes,
    the Node ids will be added, but included as empty dicts
    """
    node_extras = {}
    node_extras_conversion = {}

    for node_id in data['export_data'].get('Node', {}):
        node_extras[node_id] = {}
        node_extras_conversion[node_id] = {}
    data.update({'node_extras': node_extras, 'node_extras_conversion': node_extras_conversion})


def migrate_v3_to_v4(metadata, data, folder, *args):  # pylint: disable=unused-argument
    """
    Migration of export files from v0.3 to v0.4

    Note concerning migration 0032 - REV. 1.0.32:
    Remove legacy workflow tables: DbWorkflow, DbWorkflowData, DbWorkflowStep
    These were (according to Antimo Marrazzo) never exported.
    """
    old_version = '0.3'
    new_version = '0.4'

    verify_metadata_version(metadata, old_version)
    update_metadata(metadata, new_version)

    # Apply migrations in correct sequential order
    migration_base_data_plugin_type_string(data)
    migration_process_type(metadata, data)
    migration_code_sub_class_of_data(data)
    migration_add_node_uuid_unique_constraint(data)
    migration_migrate_builtin_calculations(data)
    migration_provenance_redesign(data)
    migration_dbgroup_name_to_label_type_to_type_string(metadata, data)
    migration_dbgroup_type_string_change_content(data)
    migration_calc_job_option_attribute_keys(data)
    migration_move_data_within_node_module(data)
    migration_trajectory_symbols_to_attribute(data, folder)
    migration_remove_node_prefix(data)
    migration_rename_parameter_data_to_dict(data)
    migration_dbnode_type_to_dbnode_node_type(metadata, data)
    migration_remove_dbcomputer_enabled(metadata, data)
    migration_replace_text_field_with_json_field(data)

    # Add Node Extras
    add_extras(data)

    # Update metadata.json with the new Log and Comment entities
    new_entities = {
        'Log': {
            'uuid': {},
            'time': {
                'convert_type': 'date'
            },
            'loggername': {},
            'levelname': {},
            'message': {},
            'metadata': {},
            'dbnode': {
                'related_name': 'dblogs',
                'requires': 'Node'
            }
        },
        'Comment': {
            'uuid': {},
            'ctime': {
                'convert_type': 'date'
            },
            'mtime': {
                'convert_type': 'date'
            },
            'content': {},
            'dbnode': {
                'related_name': 'dbcomments',
                'requires': 'Node'
            },
            'user': {
                'related_name': 'dbcomments',
                'requires': 'User'
            }
        }
    }
    metadata['all_fields_info'].update(new_entities)
    metadata['unique_identifiers'].update({'Log': 'uuid', 'Comment': 'uuid'})



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the `run` functions."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.backends.testbase import AiidaTestCase
from aiida.backends.tests.utils.processes import DummyProcess
from aiida.engine import run, run_get_node
from aiida.orm import Int, Str, ProcessNode


class TestRun(AiidaTestCase):
    """Tests for the `run` functions."""

    @staticmethod
    def test_run():
        """Test the `run` function."""
        inputs = {'a': Int(2), 'b': Str('test')}
        run(DummyProcess, **inputs)

    def test_run_get_node(self):
        """Test the `run_get_node` function."""
        inputs = {'a': Int(2), 'b': Str('test')}
        result, node = run_get_node(DummyProcess, **inputs)  # pylint: disable=unused-variable
        self.assertIsInstance(node, ProcessNode)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,too-few-public-methods
"""
Invalidating node hash - User should rehash nodes for caching
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

# Remove when https://github.com/PyCQA/pylint/issues/1931 is fixed
# pylint: disable=no-name-in-module,import-error
from django.db import migrations
from aiida.backends.djsite.db.migrations import upgrade_schema_version
from aiida.cmdline.utils.echo import echo_warning

REVISION = '1.0.39'
DOWN_REVISION = '1.0.38'

# Currently valid hash key
_HASH_EXTRA_KEY = '_aiida_hash'


def notify_user(apps, schema_editor):  # pylint: disable=unused-argument
    echo_warning('Invalidating all the hashes of all the nodes. Please run verdi rehash', bold=True)


class Migration(migrations.Migration):
    """Invalidating node hash - User should rehash nodes for caching"""

    dependencies = [
        ('db', '0038_data_migration_legacy_job_calculations'),
    ]

    operations = [
        migrations.RunPython(notify_user, reverse_code=notify_user),
        migrations.RunSQL(
            """UPDATE db_dbnode SET extras = extras #- '{""" + _HASH_EXTRA_KEY + """}'::text[];""",
            reverse_sql="""UPDATE db_dbnode SET extras = extras #- '{""" + _HASH_EXTRA_KEY + """}'::text[];"""
        ),
        upgrade_schema_version(REVISION, DOWN_REVISION)
    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi config` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments
from aiida.cmdline.utils import echo


@verdi.command('config')
@arguments.CONFIG_OPTION(metavar='OPTION_NAME')
@click.argument('value', metavar='OPTION_VALUE', required=False)
@click.option('--global', 'globally', is_flag=True, help='Apply the option configuration wide.')
@click.option('--unset', is_flag=True, help='Remove the line matching the option name from the config file.')
@click.pass_context
def verdi_config(ctx, option, value, globally, unset):
    """Configure profile-specific or global AiiDA options."""
    config = ctx.obj.config
    profile = ctx.obj.profile

    if option.global_only:
        globally = True

    # Define the string that determines the scope: for specific profile or globally
    scope = profile.name if (not globally and profile) else None
    scope_text = 'for {}'.format(profile.name) if (not globally and profile) else 'globally'

    # Unset the specified option
    if unset:
        config.unset_option(option.name, scope=scope)
        config.store()
        echo.echo_success('{} unset {}'.format(option.name, scope_text))

    # Get the specified option
    elif value is None:
        option_value = config.get_option(option.name, scope=scope, default=False)
        if option_value:
            echo.echo('{}'.format(option_value))

    # Set the specified option
    else:
        config.set_option(option.name, value, scope=scope)
        config.store()
        echo.echo_success('{} set to {} {}'.format(option.name, value, scope_text))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Whenever the requirements in ../setup.json are updated, run
also this script to update the requirements for Read the Docs.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import json
import click


@click.command()
@click.option('--pre-commit', is_flag=True)
def update_req_for_rtd(pre_commit):
    """Update the separate requirements file for Read the Docs"""
    docs_dir = os.path.abspath(os.path.dirname(__file__))
    root_dir = os.path.join(docs_dir, os.pardir)

    with open(os.path.join(root_dir, 'setup.json'), 'r') as info:
        setup_json = json.load(info)

    extras = setup_json['extras_require']
    reqs = set(extras['testing'] + extras['docs'] + extras['rest'] + extras['atomic_tools'] +
               # To avoid that it requires also the postgres libraries
               [p for p in setup_json['install_requires'] if not p.startswith('psycopg2')])
    reqs_str = '\n'.join(sorted(reqs))

    basename = 'requirements_for_rtd.txt'

    # pylint: disable=bad-continuation
    with open(os.path.join(docs_dir, basename), 'w') as reqs_file:
        reqs_file.write(reqs_str)

    click.echo("File '{}' written.".format(basename))

    if pre_commit:
        msg = 'Some requirements for Read the Docs have changed, {}'
        local_help = 'please add the changes and commit again'
        travis_help = 'please run aiida/docs/update_req_for_rtd.py locally and commit the changes it makes'
        help_msg = msg.format(travis_help if os.environ.get('TRAVIS') else local_help)
        click.echo(help_msg, err=True)


if __name__ == '__main__':
    update_req_for_rtd()  # pylint: disable=no-value-for-parameter
